import numpy as nm
import copy
import os.path as osp
from sfepy.mechanics.matcoefs import stiffness_from_lame
from sfepy.base.base import Struct, output
from porodyn_engine import post_process
from sfepy.linalg.utils import dot_sequences

hyperela = {
    'step': 0,
    'step_in_cache': -1,
    'cache': {},
    'state': {'u': None, 'du': None, 'du1': None,
              'p': None, 'dp': None},
}


def initial_problem(pb):
    from sfepy.base.conf import ProblemConf, get_standard_keywords
    from sfepy.discrete import Problem

    out_pref = output.get_output_prefix()
    output.set_output_prefix(out_pref + ' <init pb>:')

    required, other = get_standard_keywords()

    cdict = pb.conf.funmod.define()

    cdict.update({
        'options': {},
        'ebcs': {
            'Fixed_Left_u': ('Left', {'u.0': 0.0}),
            'Fixed_Right_u': ('Right', {'u.0': 0.0}),
            'Fixed_Bottom_u': ('Bottom', {'u.1': 0.0}),
            'Fixed_Top_p': ('Top_l', {'p.0': 0.0}),
        },
        'variables': {
            'u': ('unknown field', 'displacement'),
            'v': ('test field', 'displacement', 'u'),
            'p': ('unknown field', 'pressure'),
            'q': ('test field', 'pressure', 'p'),
        },
        'equations': {
            'eq1': """dw_volume_dot.5.Omega(mix.rho0, v, u)
                    = dw_surface_ltr.5.Top_r(solid.fc, v)""",
            'eq2': """dw_volume_dot.5.Omega(fluid.FKf, q, p)
                    + dw_v_dot_grad_s.5.Omega(fluid.rhoK, u, q) = 0"""
        },
    })

    funmod = Struct(__file__=pb.conf.__file__)
    conf = ProblemConf.from_dict(cdict, funmod, required, other)
    init_pb = Problem.from_conf(conf)
    init_pb.time_update()
    res = init_pb.solve().get_parts()

    output.set_output_prefix(out_pref)
    return res['u'], res['p']


def incremental_algorithm(pb):
    params = pb.conf.params
    dt, t_end = params['dt'], params['t_end']
    dim = params['dim']
    st_nd_u = params.get('u_store_node')
    st_nd_p = params.get('p_store_node')

    hist = {'p': [], 'u': []}
    hist['u'].append(st_nd_u)
    hist['p'].append(st_nd_p)

    timehist = nm.arange(0, t_end, dt) + dt
    pbvars = pb.get_variables()
    pb.domain.mesh.coors_act = pb.domain.mesh.coors.copy()

    state = hyperela['state']

    for it, time in enumerate(timehist):
        print('>>> step %d:' % it)
        hyperela['step'] = it

        if it == 0:
            out = []

            aux = pbvars['u'].field.get_coor()
            state['u'] = nm.zeros_like(aux)
            state['du'] = nm.zeros_like(aux)
            aux = pbvars['p'].field.get_coor()[:, 0].squeeze()
            state['p'] = nm.zeros_like(aux)
            state['dp'] = nm.zeros_like(aux)

            nnod = aux.shape[0]
            if st_nd_u is not None:
                hist['u'].append(state['u'][st_nd_u, dim - 1])
            if st_nd_p is not None:
                hist['p'].append(state['p'][st_nd_p])

        du = state['du']

        pbvars['P'].set_data(state['p'])
        pbvars['dU'].set_data(state['du'])

        yield pb, out

        result = out[-1][1].get_parts()
        du = result['u']
        dp = result['p']

        state['du'][:] = du.reshape(state['du'].shape)
        state['dp'][:] = dp
        state['u'][:] += du.reshape(state['u'].shape)
        state['p'][:] += dp

        tout = {}
        tout['u'] = Struct(name='output_data',
                           mode='vertex',
                           data=state['u'][:nnod, :],
                           dofs=None)
        tout['p'] = Struct(name='output_data',
                           mode='vertex',
                           data=state['p'].reshape((nnod, 1)),
                           dofs=None)

        out_fn = osp.join(pb.output_dir, '%s_%03d.vtk'
                          % (osp.split(pb.domain.mesh.name)[1], it + 1))
        pb.domain.mesh.write(out_fn, out=tout)

        if st_nd_u is not None:
            hist['u'].append(state['u'][st_nd_u, dim - 1])
        if st_nd_p is not None:
            hist['p'].append(state['p'][st_nd_p])

        yield None
        print('<<< step %d finished' % it)

    # postprocess
    post_process(timehist, hist, 'lin')


delta_delta = {
    2: nm.array([[2, 0, 0, 0],
                 [0, 1, 1, 0],
                 [0, 1, 1, 0],
                 [0, 0, 0, 2]])
}

one_one = {
    2: nm.array([[1, 0, 0, 1],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [1, 0, 0, 1]])
}


def mat_fce(mode, coors, term, pb):
    if not (mode == 'qp'):
        return

    print('### material function - step %f' % hyperela['step'])

    cache = hyperela['cache']
    # clear chache
    if hyperela['step'] != hyperela['step_in_cache']:
        for ii in cache.keys():
            del(cache[ii])

    cache_key = (hyperela['step'], term.region.name,
                 term.integral.name, term.integration)

    if cache_key not in cache:
        npts = coors.shape[0]

        state = pb.create_variables(['Um', 'Pm'])
        state_u = state['Um']
        state_p = state['Pm']

        n_el, n_qp, dim, n_en, n_c = \
            state_u.get_data_shape(term.integral, term.integration,
                                   term.region.name)
        dim2 = dim**2

        state_u.set_data(hyperela['state']['u'])
        # strain_u = state_u.evaluate(mode='cauchy_strain',
        grad_u = state_u.evaluate(mode='grad',
                                  region=term.region,
                                  integral=term.integral,
                                  integration=term.integration)
        grad_u = grad_u.reshape((n_el, n_qp, dim2, 1))
        state_p.set_data(hyperela['state']['p'])
        p_qp = state_p.evaluate(mode='val',
                                region=term.region,
                                integral=term.integral,
                                integration=term.integration)

        par_s = pb.conf.material_params['solid']

        mat_mu = nm.ones((n_el, n_qp, 1, 1)) * par_s['mu']
        mat_lam = nm.ones((n_el, n_qp, 1, 1)) * par_s['lam']

        tanmod = mat_mu * delta_delta[dim] + mat_lam * one_one[dim]

        stress = dot_sequences(tanmod, grad_u)
        stress_s = nm.empty((n_el, n_qp, 3, 1), dtype=nm.float64)
        stress_s[:, :, 0, 0] = stress[:, :, 0, 0]
        stress_s[:, :, 1, 0] = stress[:, :, 3, 0]
        stress_s[:, :, 2, 0] = (stress[:, :, 1, 0] + stress[:, :, 2, 0]) * 0.5
        stress_s[:, :, 0, 0] -= p_qp[:, :, 0, 0]
        stress_s[:, :, 1, 0] -= p_qp[:, :, 0, 0]

        cache[cache_key] = {
            'A': tanmod.reshape((npts, dim2, dim2)),
            'S': stress_s.reshape(npts, 3, 1),
        }

        hyperela['step_in_cache'] = hyperela['step']
        print('  computed')

    else:
        print('  cached')

    return copy.deepcopy(cache[cache_key])


def define():
    params = {
        'mesh_file': 'rect_16x16.vtk',
        'mat_store_elem': 75,   # element for which material data are stored
        'u_store_node': 272,    # node for which displacement is stored
        'p_store_node': 144,    # node for which pressure is stored
        'dim': 2,               # problem dimension
        'dt': 0.01,             # time step
        't_end': 2.0,           # end time
        'force': 4e6,           # applied force
        'save_step': True,      # save results in each time step?
        'init_mode': False,     # calculate initial state?
    }

    material_params = {
        'param': {
            'B': nm.eye(params['dim']),
            'g': 9.81,      # gravitational acceleration
        },
        'solid': {
            'Phi': 0.58,    # volume fraction
            'lam': 8.4e6,   # Lame coefficient
            'mu': 5.6e6,    # Lame coefficient
            'rho': 2700,    # density
        },
        'fluid': {
            'kappa': 1e-1,  # permeability parameter
            'beta': 0.8,    # permeability parameter
            'rho': 1000,    # density
            'Kf': 2.2e10,   # bulk modulus
        },
    }

    regions = {
        'Omega': 'all',
        'Left': ('vertices in (x < 0.001)', 'facet'),
        'Right': ('vertices in (x > 9.999)', 'facet'),
        'Bottom': ('vertices in (y < 0.001)', 'facet'),
        'Top_r': ('vertices in (y > 9.999) & (x > 4.999)', 'facet'),
        'Top_l': ('vertices in (y > 9.999) & (x < 5.001)', 'facet'),
    }

    ebcs = {
        'Fixed_Left_u': ('Left', {'u.0': 0.0}),
        'Fixed_Right_u': ('Right', {'u.0': 0.0}),
        'Fixed_Bottom_u': ('Bottom', {'u.1': 0.0}),
        'Fixed_Top_p': ('Top_l', {'p.0': 0.0}),
    }

    ###############################################
    par = material_params['param']
    par_s = material_params['solid']
    par_f = material_params['fluid']
    dim = params['dim']

    rho0 = par_s['Phi'] * par_s['rho'] + (1. - par_s['Phi']) * par_f['rho']
    parK = par_f['kappa'] / (par['g'] * par_f['rho'])

    materials = {
        'mat': 'mat_fce',
        'solid': ({
            'D': stiffness_from_lame(dim, par_s['lam'], par_s['mu']),
            'fc': nm.array([0, -params['force']]).reshape((2, 1)),
        },),
        'fluid': ({
            'rhoK': nm.eye(dim) * parK * par_f['rho'],
            'K': nm.eye(dim) * parK,
            'FKf': (1. - par_s['Phi']) / par_f['Kf'],
        },),
        'mix': ({
            'rho0': rho0,
        },),
    }

    functions = {
        'mat_fce': (lambda ts, coors, mode=None, term=None,
            problem=None, **kwargs: mat_fce(mode, coors, term, problem),),
    }

    options = {
        'output_dir': 'output',
        'parametric_hook': 'incremental_algorithm',
    }

    variables = {
        'u': ('unknown field', 'displacement'),
        'v': ('test field', 'displacement', 'u'),
        'p': ('unknown field', 'pressure'),
        'q': ('test field', 'pressure', 'p'),
        'U': ('parameter field', 'displacement', '(set-to-None)'),
        'dU': ('parameter field', 'displacement', '(set-to-None)'),
        'Um': ('parameter field', 'displacement', '(set-to-None)'),
        'P': ('parameter field', 'pressure', '(set-to-None)'),
        'Pm': ('parameter field', 'pressure', '(set-to-None)'),
    }

    filename_mesh = params['mesh_file']

    dt = params['dt']
    equations = {
        'balance_of_forces':
            """
          %e * dw_volume_dot.5.Omega(mix.rho0, v, u)
             + dw_nonsym_elastic.5.Omega(mat.A, v, u)
             - dw_stokes.5.Omega(v, p)
             = dw_surface_ltr.5.Top_r(solid.fc, v)
             - dw_lin_prestress.5.Omega(mat.S, v)
          %e * dw_volume_dot.5.Omega(mix.rho0, v, dU)""" % (dt**-2, dt**-2),
        'fluid':
            """
          %e * dw_v_dot_grad_s.5.Omega(fluid.rhoK, u, q)
             + dw_volume_dot.5.Omega(fluid.FKf, q, p)
             + dw_stokes.5.Omega(u, q)
        + %e * dw_diffusion.5.Omega(fluid.K, q, p)
             =
          %e * dw_v_dot_grad_s.5.Omega(fluid.rhoK, dU, q)
        - %e * dw_diffusion.5.Omega(fluid.K, q, P)""" % (1./dt, dt, 1./dt, dt),
    }

    fields = {
        'displacement': ('real', 'vector', 'Omega', 2),
        'pressure': ('real', 'scalar', 'Omega', 1),
    }

    solvers = {
        'ls': ('ls.scipy_direct', {}),
        'newton': ('nls.newton', {'eps_a': 1e-6, 'eps_r': 1e-3}),
    }

    return locals()
